Attribute VB_Name = "mdGlyphOutline"
' ***************************************************************************************
'     NAME: Modul GlyphOutline
'     DESC: Modul zum Handling der GlyphOutline-API der Windows gdi32.dll
'     DESC: zum ermittlen der Zeichenumrandungen von TrueType Zeichenstzen
' ***************************************************************************************
'
'     AUTHOR:  Stefan Maag
'     CREATE:  12/01/2003
'     CHANGE:
'     COPY:    Der Code geht auf eine Ausschreibung bei Activevb und ein
'     COPY:    Beispiel von Klaus Langbein zurck. Die Idee dazu hatte Robert Christ
'     COPY:    Dieser Code ist frei verfgbar. Die Verwendung des Codes
'     COPY:    oder Codeteilen erfolgt immer auf eigenes Risiko und Verantwortung

' ===========================================================================
'  REM: Quellenangaben:
' ===========================================================================
'  REM:
'  REM:  1: www.activevb.de Beispielcode zu GlyphOutline von Klaus Langbein
'  REM:  2: MSDN Library Juli 2001, Microsoft
'  REM:  3: Das Beispiel in C zu GetGlyphOutline aus der MSDN von Ron Gery
'  REM:     Microsoft Developer Network Technology Group reated: July 10, 1992
'  REM:
' ===========================================================================
'  REM: Weitere bentigte Dateien:
' ===========================================================================
'
'  REM: keine

Option Explicit

Public Enum TT_GlyphFormat
' DESC: Konstanten fr den fuFormat-Paramter der GlyphOutline-Funktion
   GGO_BITMAP = 1&     ' DESC: liefert den Zeichenumriss als Bitmap
   GGO_METRICS = 0&    ' DESC: liefert nur die GlyphMetrics Struktur zurck Buffer wird ignoriertEnd Enum
   GGO_NATIVE = 2&     ' DESC: liefert den Zeichenumriss als Array von Eckpunkten, Polygon
End Enum

 
Public Enum TT_CurveType
' DESC: Typ der Punkte im Buffer
   TT_PRIM_LINE = 1&
   TT_PRIM_QSPLINE = 2&
   TT_POLYGON_TYPE = 24&
End Enum

'Public Const GGO_BITMAP = 1      ' DESC: liefert den Zeichenumriss als Bitmap
'Public Const GGO_METRICS = 0     ' DESC: liefert nur die GlyphMetrics Struktur zurck Buffer wird ignoriert
'Public Const GGO_NATIVE = 2      ' DESC: liefert den Zeichenumriss als Array von Eckpunkten, Polygon

'Public Const TT_POLYGON_TYPE = 24
'Public Const TT_PRIM_LINE = 1
'Public Const TT_PRIM_QSPLINE = 2

'  Bei dem Fixed Datentyp handelt es sich um eine 32 Bit Festkommawert
'  mit 16 Bit Ganzzahlanteil und 16 Bit Nachkommanateil. Dies ist
'  jedoch etwas umstndlich zu handhaben. Leichter zu verstehen ist
'  es, wenn man diesen Wert als normalen Long betrachtet, der um eine
'  bessere Auflsung der Koordinaten zu bekommen, einfach mit einem
'  Faktor multipliziert ist.
'  Hier ist dieser Faktor 65536, also um 16 Bits nach links geschoben.
'
'  Wir werden diesen Datentyp hier nicht verwenden, sondern die Werte
'  immer als Long betrachten und gegebenfalls durch 65536 teilen und
'  das Ergebnis in einen Gleitpunkt Datentyp (Single/Double) speichern.
'  Dies spart uns viel rger!

Type FIXED
    Fract As Integer   ' Nachkommanteil
    Value As Integer   ' Ganzzahlanteil
End Type

Type POINTFX
' DESC: Punkt in der Festkommadefinition der GlyphOutline Funktion
    x As FIXED
    y As FIXED
End Type

Type PointAPI
' DESC: Punkt im Format des Windows API, diesen verwenden wir auch als Ersatz fr POINTFX
    x As Long
    y As Long
End Type

Type PointShort
' DESC: Punkt in der Definition mit 16 Bit Integer-Koordinaten
    x As Integer
    y As Integer
End Type

Type PointSingle
' DESC:  Punkt in der Definition mit (32 Bit) Single-Koordinaten
    x As Single
    y As Single
End Type

Type GLYPHMETRICS
    gmBlackBoxX As Long    ' Breite der Bitmaps bei fuFormat = CGO_BITMAP
    gmBlackBoxY As Long    ' Hhe der Bitmaps bei fuFormat = CGO_BITMAP
    gmptGlyphOrigin As PointAPI
    gmCellIncX As Integer
    gmCellIncY As Integer
End Type


Type MAT2   ' DESC: die fr VB-Programme leichter zu handhabende MAT2 Struktur mit Long
    eM11 As Long
    eM12 As Long
    eM21 As Long
    eM22 As Long
End Type

'Type MAT2          ' DESC: Originaldefintion der MAT2 Struktur mit FIXED Datentyp
'    eM11 As FIXED
'    eM12 As FIXED
'    eM21 As FIXED
'    eM22 As FIXED
'End Type

' DESC: MAT2 ist die Definition fr die Transformationsmatrix.
' DESC: Diese Matrix gibt die Grssenverhltnisse und die Ausrichtung,
' DESC: Drehung des erzeugten Zeichenumrisses an.
' DESC:
' DESC: Wir verwenden hier aber nicht die Originaldefinition mit den
' DESC: Werten als FIXED Datentyp sondern wir verwenden wegen besseren
' DESC: Handlings einen einfachen Long. Als Ausgangswert zur Berechung
' DESC: der em-Werte verwenden wir wieder einen Gleitkommawert und
' DESC: multiplizieren diesen mit 65536, speichern diesen in einem Long
' DESC: und erhalten so unseren Festkommawert

' DESC: Die Matrix hat folgendes Aussehen

' DESC: em11 em12
' DESC: em21 em22

Type TTPOLYGONHEADER
    cb As Long             ' Summe der Bytes fr Header und Curves
    dwType As Long         ' immer TT_POLYGON_TYPE.
    ' pfxStart As POINTFX   ' erster Punkt als POINTFX Struktur: nicht verwenden
    pfxStart As PointAPI      ' erster Punkt, Punktkoordinaten sind Long
End Type

Type TTPOLYCURVE
    wType As Integer    ' Typ: Polygon oder Bezier
    cpfx As Integer     ' Anzahl der folgenden Punkte incl. dem Startpunkt
    ' apfx As POINTFX    ' Startpunkt der Kurve als POINTFX Struktur: nicht verwenden
    apfx As PointAPI       ' Startpunkt der Kurve, Punktkoordinaten sind Long
End Type

Public Declare Function GetGlyphOutline Lib "gdi32" Alias "GetGlyphOutlineA" _
                  (ByVal hdc As Long, ByVal uChar As Long, _
                   ByVal fuFormat As Long, lpgm As GLYPHMETRICS, _
                   ByVal cbBuffer As Long, lpBuffer As Any, lpmat2 As MAT2) As Long

' Die GlyphOutline Funktion ermittelt die Zeichenumrisse (Outline)

'DWORD GetGlyphOutline(
'  HDC hdc,             // handle to DC
'  UINT uChar,          // character to query
'  UINT uFormat,        // data format
'  LPGLYPHMETRICS lpgm, // glyph metrics
'  DWORD cbBuffer,      // size of data buffer
'  LPVOID lpvBuffer,    // data buffer
'  CONST MAT2 *lpmat2   // transformation matrix
');

Public Const FixedFaktor = 65536    ' DESC: Faktor zur Berechnung des FIXED Datenformates (:= Shift_16)

Private Declare Sub MoveMemory Lib "kernel32" Alias "RtlMoveMemory" _
                    (hpvDest As Any, hpvSource As Any, ByVal Bytes As Long)

Private Declare Sub MoveMemoryVal Lib "kernel32" Alias "RtlMoveMemory" _
                    (hpvDest As Any, ByVal hpvSource As Long, ByVal Bytes As Long)

Private Declare Sub PeekPoint Lib "msvbvm60.dll" Alias "GetMem8" (Ptr As Any, RetVal As PointAPI)

Public Function GetIdentityMatrix() As MAT2
' ===========================================================================
'  NAME:    GetIdentityMatrix
'  DESC:    Erzeugt die IdentityMatrix, diese ist im Prinzip eine 1 und
'  DESC:    verndert das Zeichen weder in Gre noch Richtung
'  DESC:    man bekommt einen unmaipulierten Glyph
'  RETURN:  IdentityMatrix As MAT2
' ===========================================================================
   With GetIdentityMatrix
      .eM11 = 1 * FixedFaktor
      .eM12 = 0
      .eM21 = 0
      .eM22 = 1 * FixedFaktor
   End With
End Function

Public Function GetShearMatrix() As MAT2
' ===========================================================================
'  NAME:    GetShearMatrix
'  DESC:    Erzeugt die horizontale ShearMatrix, Italic Font Simualtion
'  RETURN:  ShearMatrix As MAT2
' ===========================================================================
   With GetShearMatrix
      .eM11 = 1 * FixedFaktor
      .eM12 = 0
      .eM21 = 0.25 * FixedFaktor
      .eM22 = 1 * FixedFaktor
   End With
End Function

Public Function GetRotationMatrix(Angle As Double) As MAT2
' ===========================================================================
'  NAME:    GetRotationMatrix
'  DESC:    Erzeugt eine Rotationsmatrix basierend auf dem Winkel in Grad
'  RETURN:  Rotationsmatrix As MAT2
' ===========================================================================
   Const Pi = 3.14159265358979
   Dim angl As Double
   
   angl = Angle * Pi / 180
   With GetRotationMatrix
      .eM11 = (Cos(angl)) * CDbl(FixedFaktor)
      .eM12 = Sin(angl) * CDbl(FixedFaktor)
      .eM21 = -.eM12
      .eM22 = .eM11
   End With
End Function

Public Function GetStrechMatrix(ByVal StrechX As Single, ByVal StrechY As Single) As MAT2
' ===========================================================================
'  NAME:    GetStrechMatrix
'  DESC:    Erzeugt eine Strechmatrix basierend auf den Steckfaktoren fr x/y
'  DESC:    verndert das Zeichen in der Gre
'  RETURN:  StrechMatrix As MAT2
' ===========================================================================
   With GetStrechMatrix
      .eM11 = StrechX * FixedFaktor
      .eM12 = 0
      .eM21 = 0
      .eM22 = StrechY * FixedFaktor
   End With
End Function

Public Function GetOutline(Buffer() As Long, ByVal hdc As Long, ByVal CharASCII As Long, ByVal fuFormat As TT_GlyphFormat, _
                           metr As GLYPHMETRICS, Matrix As MAT2) As Long
' ===========================================================================
'  NAME:    GetOutline
'  DESC:    speichert die Punkte der Outline-Kurve in Buffer
'  RETURN:  Anzahl der Bytes in Buffer (nicht der Eintrge)
' ===========================================================================
   
   Dim ret As Long
   Dim ByteSize As Long       ' Anzahl der bentigten Bytes im Buffer
   Dim BufSize As Long        ' Eintrge im Buffer ( := Ubound(Buffer()) )
   Dim Ptr As Long            ' Pointerwert
   
   ' Beim ersten Aufruf (ByteSize=0) wird die bentigte Lnge des Buffers in
   ' Bytes zurckgegen. Oder falls Funktion fehlschlgt der Fehlercode (<0)
   ret = GetGlyphOutline(hdc, CharASCII, fuFormat, metr, ByteSize, ByVal Ptr, Matrix)
   
   If ret > 0 Then
        ByteSize = ret
        BufSize = (ret / 4) - 1  ' /4, da es sich um einen LongBuffer und nicht um Bytes handelt
        ' -1, da von 0 ab dimensioniert wird und BufSize die Anzahl der Eintrge darstellt
   Else
        GetOutline = ret       ' Fehlercode zurckgeben
        Exit Function
   End If

   ReDim Buffer(BufSize) As Long    ' Buffer in der Bentigten Gre anlegen
   
   
   Ptr = VarPtr(Buffer(0))          ' Startadresse von Buffer()
   
   ' Nun beim 2ten Aufruf wird wirklich der Umriss generiert und in Buffer() gespeichert
   ret = GetGlyphOutline(hdc, CharASCII, fuFormat, metr, ByteSize, ByVal Ptr, Matrix)
   
   GetOutline = ret       ' Returncode zurckgeben
   
   If ret <= 0 Then
      MsgBox "GetGlyphOutline: Error!"
      Exit Function
   End If
   
End Function

Public Sub DrawGlyph(Buffer() As Long, pb As PictureBox, ByVal xoff As Long, ByVal yoff As Long)
' ===========================================================================
'  NAME:    DrawGlyph
'  DESC:    zeichnet den Umriss in die angegeben Picturebox
' ===========================================================================
   
   Dim I As Long
   Dim j As Long
   Dim idx As Long
   Dim UB As Long
   Dim EndPoly As Long
   Dim PtsCnt As Long
   Dim ptStart As PointAPI
   Dim x As Single
   Dim y As Single
   Dim typ As Long
   
   Dim xs() As Long
   Dim ys() As Long
   Dim xp(2) As Long
   Dim yp(2) As Long
   Dim pt() As PointAPI

   UB = UBound(Buffer())
   
   ' Alle Polygonheader mit untergeordneten PolyCurves durchlaufen
   Do
      ' --------------------------------------------------------------
      '  Daten der TTPOLGONHEADER - Struktur ausarbeiten
      ' --------------------------------------------------------------
      
      ' Berechnung des letzten Eintrags des Polygonzugs
      ' EndPoly = Lnge_Polygon_in_Bytes /4 + aktueller_BufferIndex
      EndPoly = Buffer(idx) \ 4 + idx
      
      If Buffer(idx + 1) <> TT_POLYGON_TYPE Then
         MsgBox "Fehler Polygon zeichnen: Kurve ist kein Polygonzug"
         Exit Sub
      End If
      
      ' Startpunkt des Polygonzuges
      ptStart.x = Buffer(idx + 2)
      ptStart.y = Buffer(idx + 3)
      
      x = ptStart.x / FixedFaktor + xoff
      y = yoff - ptStart.y / FixedFaktor
      
      pb.PSet (x, y), 0
      idx = idx + 4
      
      Do
      ' --------------------------------------------------------------
      ' Es beginnen die TTPOLYCURVE-Strukturen
      ' --------------------------------------------------------------
         
         ' Alle Kurven des Polygons durchlaufen
         PtsCnt = Buffer(idx) \ 65536   ' HiWord = Anzahl der folgenden Punkte
         typ = Buffer(idx) And 65535
         
         idx = idx + 1              ' idx auf X-Koordinate des Ersten der PtsCnt-Punkte
         
         Select Case typ

         Case TT_PRIM_LINE
            
            For I = 1 To PtsCnt
               x = Buffer(idx) / FixedFaktor + xoff
               y = yoff - Buffer(idx + 1) / FixedFaktor
               pb.Line -(x, y)
               idx = idx + 2
               ' MsgBox idx
            Next
         
         Case TT_PRIM_QSPLINE
         
            ReDim xs(1 To PtsCnt)
            ReDim ys(1 To PtsCnt)
                
            For I = 1 To PtsCnt
               xs(I) = xoff + Buffer(idx) / FixedFaktor
               ys(I) = yoff - Buffer(idx + 1) / FixedFaktor
               idx = idx + 2
            Next I
                
            For I = 1 To PtsCnt - 1
                
               xp(0) = pb.CurrentX
               yp(0) = pb.CurrentY
                                       
               xp(1) = xs(I)
               yp(1) = ys(I)
               
               Select Case PtsCnt - I
               Case 0
                   
               Case 1
                  xp(2) = xs(I + 1)
                  yp(2) = ys(I + 1)
               Case Else
                  xp(2) = xp(1) + (xs(I + 1) - xp(1)) / 2
                  yp(2) = yp(1) + (ys(I + 1) - yp(1)) / 2
               End Select
                    

               pb.CurrentX = xp(0)
               pb.CurrentY = yp(0)
                    
                    
               Call Qspline(30, xp(), yp(), pt())
                    
               For j = 0 To UBound(pt)
                   pb.Line -(pt(j).x, pt(j).y)
               Next j
                                        
            Next I
         
         End Select
      ' --------------------------------------------------------------
      ' Alle TTPOLYCURVE-Strukturen diese Polygonzuges abgeschlossen?
      ' --------------------------------------------------------------
      Loop Until idx >= (EndPoly)
      
      ' nun noch die Kurve zum Ausgangspunkt schliessen
      pb.Line -(ptStart.x / FixedFaktor + xoff, yoff - ptStart.y / FixedFaktor)
      
   ' --------------------------------------------------------------
   ' Alle TTPOLYGONHEADER-Strukturen des Glyph abgearbeitet?
   ' --------------------------------------------------------------
   Loop Until idx >= UB ' Ende des Buffers erreicht?
   
End Sub

Sub Qspline(ByVal n As Long, ByRef x() As Long, ByRef y() As Long, ByRef ptOut() As PointAPI)
    
    Dim I As Long
    Dim t As Double
    Dim tstep As Double
    ReDim ptOut(0 To n)
    
    tstep = 1 / (n)
    
    For I = 0 To n
    
        t = I * tstep
        ptOut(I).x = (x(0) - 2 * x(1) + x(2)) * t ^ 2 + (2 * x(1) - 2 * x(0)) * t ^ 1 + x(0)
        ptOut(I).y = (y(0) - 2 * y(1) + y(2)) * t ^ 2 + (2 * y(1) - 2 * y(0)) * t ^ 1 + y(0)
         
    Next I

End Sub

'Private Sub PrintArray(ar() As Long)
'   Dim I As Long
'   Debug.Print
'   For I = LBound(ar()) To UBound(ar())
'      Debug.Print I, ar(I)
'   Next
'End Sub
'
'Private Sub PrintPointArray(ar() As PointAPI)
'   Dim I As Long
'   Debug.Print
'   For I = LBound(ar()) To UBound(ar())
'      Debug.Print I, ar(I).x / FixedFaktor, ar(I).y / FixedFaktor
'   Next
'End Sub

